#define UNICODE
#define _UNICODE

#include "windows.h"
#include "perfc.h"
#include "perfm.h"

DWORD WINAPI MonitorThread(LPVOID);
LRESULT CALLBACK PerfcWndProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK SettingsDlgProc(HWND, UINT, WPARAM, LPARAM);

HINSTANCE hInst;
PERFC_DATA MemData, DiskData;

//-----------------------------------------------------------------
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst, 
                     LPSTR lpCmdLine, int nCmdShow)
{
    WNDCLASS wndClass;
    DWORD dwThreadID;
    MSG msg;

    hInst = hInstance;

    memset(&wndClass, 0, sizeof(WNDCLASS));
    wndClass.lpfnWndProc = PerfcWndProc;
    wndClass.hInstance = hInst;
    wndClass.lpszClassName = TEXT("PerfcWndClass");
    RegisterClass(&wndClass);

    MemData.hWnd = CreateWindow(TEXT("PerfcWndClass"), 
		                TEXT("Perfc"), WS_POPUP, 
                                CW_USEDEFAULT, 0, CW_USEDEFAULT, 
                                0, NULL, NULL, hInst, NULL);

    // OBJECT:   "Memory" (4)
    // COUNTER:  "Available Bytes" (24)
    // INSTANCE: none

    GetCounterNumbers(TEXT("Memory"), MemData.szObject,
                      TEXT("Available Bytes"), MemData.szCounter);

    MemData.dwId           = PERFC_MEM;
    MemData.szInstance[0]  = 0;
    MemData.dwFreq         = DEF_MEM_FREQ;
    MemData.dwThrAlert     = DEF_MEM_THRESH_ALERT;
    MemData.dwThrWarn      = DEF_MEM_THRESH_WARN;
    MemData.bThrLow        = TRUE;
 
    MemData.hMonitorThread = CreateThread(NULL, 0, MonitorThread, 
                                 (PVOID)&MemData, 0, &dwThreadID);

    // OBJECT:  "LogicalDisk" (236)
    // COUNTER: "Free Megabytes" (410)
    // INSTANCE: x: (system drive, determined programatically)

    DiskData.hWnd = CreateWindow(TEXT("PerfcWndClass"), 
                                 TEXT("Perfc"), WS_POPUP, 
                                 CW_USEDEFAULT, 0, CW_USEDEFAULT, 
                                 0, NULL, NULL, hInst, NULL);

    GetCounterNumbers(TEXT("LogicalDisk"), DiskData.szObject,
                      TEXT("Free Megabytes"), DiskData.szCounter);

    GetSystemDirectory(DiskData.szInstance, MAX_PATH);
    DiskData.szInstance[2] = 0;
    DiskData.dwId           = PERFC_DISK;
    DiskData.dwFreq         = DEF_DISK_FREQ;
    DiskData.dwThrAlert     = DEF_DISK_THRESH_ALERT;
    DiskData.dwThrWarn      = DEF_DISK_THRESH_WARN;
    DiskData.bThrLow        = TRUE;
 
    DiskData.hMonitorThread = CreateThread(NULL, 0, MonitorThread,
                              (PVOID)&DiskData, 0, &dwThreadID);

    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return TRUE;    
} // WinMain

//-----------------------------------------------------------------
DWORD WINAPI MonitorThread(LPVOID lpParam)
{
    NOTIFYICONDATA notifyIconData;
    HICON hIcon[3];
    TCHAR szTip[3][64];
    DWORD dwCntNum, dwValue = 0, dwIndex = PERFC_OK, dwType, i;
    PPERFC_DATA p = (PPERFC_DATA)lpParam;

    notifyIconData.cbSize = sizeof(NOTIFYICONDATA);
    notifyIconData.hWnd = p->hWnd;
    notifyIconData.uID = 0;
    notifyIconData.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
    notifyIconData.uCallbackMessage = WM_APP + p->dwId;

    hIcon[PERFC_OK] = LoadIcon(hInst,
        MAKEINTRESOURCE(p->dwId + PERFC_ICON_BASE + PERFC_OK));
    hIcon[PERFC_WARN] = LoadIcon(hInst,
        MAKEINTRESOURCE(p->dwId + PERFC_ICON_BASE + PERFC_WARN));
    hIcon[PERFC_ALERT] = LoadIcon(hInst,
        MAKEINTRESOURCE(p->dwId + PERFC_ICON_BASE + PERFC_ALERT));

    i = p->dwId + PERFC_STR_BASE;
    LoadString(hInst, i + PERFC_OK, szTip[PERFC_OK], 64);
    LoadString(hInst, i + PERFC_WARN, szTip[PERFC_WARN], 64);
    LoadString(hInst, i + PERFC_ALERT, szTip[PERFC_ALERT], 64);
       
    dwCntNum = _wtoi(p->szCounter);
    dwType = NIM_ADD;   // first time

    while (TRUE) {

        if (GetCounterData(p->szObject, p->szInstance, dwCntNum,
                           (LPBYTE)&dwValue, sizeof(DWORD))) {
            if (p->bThrLow) {
                // warn/alert if value LESS than threshold
                if (dwValue < p->dwThrAlert) {
                    dwIndex = PERFC_ALERT;
                } else if (dwValue < p->dwThrWarn) {
                    dwIndex = PERFC_WARN;
                } else {
                    dwIndex = PERFC_OK;
                }
            } else {
                // warn/alert if value GREATER than threshold
                if (dwValue > p->dwThrAlert) {
                    dwIndex = PERFC_ALERT;
                } else if (dwValue > p->dwThrWarn) {
                    dwIndex = PERFC_WARN;
                } else {
                    dwIndex = PERFC_OK;
                }
            }
        }
        notifyIconData.hIcon = hIcon[dwIndex];
        lstrcpy(notifyIconData.szTip, szTip[dwIndex]);

        Shell_NotifyIcon(dwType, &notifyIconData);

        dwType = NIM_MODIFY;  // subsequent times
        Sleep(p->dwFreq);
    }

    Shell_NotifyIcon(NIM_DELETE, &notifyIconData);
    return 0;

} // MonitorThread

//-----------------------------------------------------------------
LRESULT CALLBACK PerfcWndProc(HWND hwnd, UINT uMsg, WPARAM wParam,
                              LPARAM lParam)
{
    switch (uMsg) {
        case WM_APP + PERFC_MEM:
            switch (lParam) {
                case WM_LBUTTONUP:
                    DialogBoxParam(hInst, 
                          MAKEINTRESOURCE(DLG_SETTINGS), 
                          NULL, SettingsDlgProc, (LPARAM)&MemData);
            }
            break;

        case WM_APP + PERFC_DISK:
            switch (lParam) {
                case WM_LBUTTONUP:
                    DialogBoxParam(hInst, 
                          MAKEINTRESOURCE(DLG_SETTINGS), 
                          NULL, SettingsDlgProc,(LPARAM)&DiskData);
            }
            break;

        case WM_DESTROY:
            TerminateThread(MemData.hMonitorThread, 0);
            TerminateThread(DiskData.hMonitorThread, 0);
            PostQuitMessage(0);
            break;

        default:
            return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
    return 0;
} // PerfcWndProc

//-----------------------------------------------------------------
int CALLBACK SettingsDlgProc(HWND hDlg, UINT Message,WPARAM wParam,
                             LPARAM lParam)
{
    TCHAR sz[MAX_PATH];
    static PPERFC_DATA p;

    switch (Message) {
        case WM_INITDIALOG:
            p = (PPERFC_DATA)lParam;
            SetForegroundWindow(hDlg);

            LoadString(hInst, p->dwId+PERFC_LABEL, sz, 64);
            SetDlgItemText(hDlg, ID_LABEL, sz);
            SetDlgItemInt(hDlg,ID_THR_WARN,p->dwThrWarn,FALSE);
            SetDlgItemInt(hDlg,ID_THR_ALERT,p->dwThrAlert,FALSE);
            SetDlgItemInt(hDlg,ID_FREQ,p->dwFreq,FALSE); 
            if (p->bThrLow) {
                CheckRadioButton(hDlg, ID_THR_LESS, ID_THR_MORE, 
                                 ID_THR_LESS);
            } else {
                CheckRadioButton(hDlg, ID_THR_LESS, ID_THR_MORE, 
                                 ID_THR_MORE);
            }
            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                case IDOK:
                    SuspendThread(p->hMonitorThread);
                    p->dwThrWarn = GetDlgItemInt(hDlg,ID_THR_WARN,
                                         NULL,FALSE);
                    p->dwThrAlert = GetDlgItemInt(hDlg,
                                         ID_THR_ALERT,NULL,FALSE);
                    p->dwFreq = GetDlgItemInt(hDlg,ID_FREQ,
                                         NULL,FALSE);
                    p->bThrLow = IsDlgButtonChecked(hDlg, 
                                                    ID_THR_LESS);
                    ResumeThread(p->hMonitorThread);
                    EndDialog(hDlg, TRUE);
                    return TRUE;

                case IDCANCEL:
                    EndDialog(hDlg, TRUE);
                    return TRUE;
            }

        default:
            return FALSE;
    }
    return FALSE;

} // SettingsDlgProc
