// PropBox.cpp
//      Implementation of the CPropertyBox class
//      Written by Christopher Dix, 2000
#include "stdafx.h"
#include "propbox.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//////////////////////////////////////////////////////////////
// CPropertyBox 
CPropertyBox::CPropertyBox() : m_nPropertyWidth(0), 
    m_pEdit(NULL), m_pButton(NULL), m_pODButton(NULL),
    m_pList(NULL), m_pMeasureFont(NULL), m_bTabWithinBox(true)
{}

CPropertyBox::~CPropertyBox()
{
    delete m_pEdit;
    delete m_pButton;
    delete m_pODButton;
    RemoveAll();
}

BEGIN_MESSAGE_MAP(CPropertyBox, CListBox)
    //{{AFX_MSG_MAP(CPropertyBox)
    ON_CONTROL_REFLECT(LBN_SELCHANGE, OnSelchange)
    ON_WM_MEASUREITEM_REFLECT()
    ON_WM_DRAWITEM_REFLECT()
    ON_WM_SIZE()
    ON_WM_GETDLGCODE()
    ON_CONTROL_REFLECT(LBN_SETFOCUS, OnSetfocus)
    //}}AFX_MSG_MAP
    END_MESSAGE_MAP()

//////////////////////////////////////////////////////////////
// CPropertyBox message handlers
BOOL CPropertyBox::PreCreateWindow(CREATESTRUCT& cs) 
{
    cs.style |= LBS_NOTIFY | LBS_SORT | WS_TABSTOP | 
        LBS_WANTKEYBOARDINPUT | LBS_OWNERDRAWFIXED | 
        LBS_HASSTRINGS | LBS_NOINTEGRALHEIGHT;
    cs.dwExStyle |= WS_EX_CLIENTEDGE;
    return( CListBox::PreCreateWindow( cs ) );
}

void CPropertyBox::PreSubclassWindow()
{
    // Create the floating edit window.
    m_pEdit = new CRichEditCtrl();
    m_pEdit->Create( WS_CHILD | WS_TABSTOP | ES_NOHIDESEL |
        ES_AUTOHSCROLL, CRect(), this, PB_EDIT_ID );
    m_pEdit->SetFont( GetFont() );
    
    // Subclass the edit window.
    WNDPROC poldproc =  (WNDPROC)::GetWindowLong( 
        m_pEdit->GetSafeHwnd(), GWL_WNDPROC );
    SetWindowLong( m_pEdit->GetSafeHwnd(), GWL_USERDATA, 
        (LONG) poldproc );
    SetWindowLong( m_pEdit->GetSafeHwnd(), GWL_WNDPROC, 
        (LONG) ChildSubclassProc );
    
    // Create the drop down button
    m_pButton = new CButton();
    m_pButton->Create( "", WS_CHILD | BS_CENTER | BS_VCENTER | 
        BS_NOTIFY | BS_BITMAP, CRect(), this, PB_VALBUTTON_ID );
    m_pButton->SetBitmap( ::LoadBitmap( NULL, 
        MAKEINTRESOURCE( OBM_COMBO ) ) );
    
    // Create the owner drop down button
    m_pODButton = new CButton();
    m_pODButton->Create( "...", WS_CHILD | BS_CENTER | BS_VCENTER
        | BS_NOTIFY, CRect(), this, PB_ODBUTTON_ID );
    SetWindowLong( m_hWnd, GWL_USERDATA, 0 );
    SetFont( GetParent()->GetFont() );
}

void CPropertyBox::OnSelchange() 
{
    CRect rectEdit, rectButton;
    HideChildren();
    
    // Get the newly selected property
    CPropertyItem* pItem = NULL;
    pItem = GetProperty( GetCurSel() );
    if ( pItem == NULL ) return;
    
    // Reposition and update the floating child windows
    m_pEdit->SetWindowText( pItem->m_strValue );
    m_pList = (pItem->m_pLB);
    GetItemRect( GetCurSel(), rectButton );
    rectButton.right--;
    rectButton.bottom--;
    rectEdit = m_rectValue;
    if ( pItem->m_nFlags & PROP_READONLY ) return;
    
    // Handle the buttons
    if ( ( pItem->m_nFlags & PROP_DROPLIST ) ||
        ( pItem->m_nFlags & PROP_OWNERDROP ) )
    {
        rectEdit.right -= kDefButtonWidth;
        rectButton.left = rectEdit.right;
        CButton* pButton = m_pButton;
        if (pItem->m_nFlags & PROP_OWNERDROP )
            pButton = m_pODButton;
        pButton->MoveWindow( rectButton );
        pButton->InvalidateRect( NULL );
        pButton->ShowWindow( SW_SHOW );
    }
    m_pEdit->MoveWindow( rectEdit );
    m_pEdit->ShowWindow( SW_SHOW );
    
    // If the PROP_EDIT flag is active, the edit window is too
    SetWindowLong( m_hWnd, GWL_USERDATA, (pItem->m_nFlags & 
		PROP_EDIT ? 0 : kAllowEdit) );
}

void CPropertyBox::OnSetfocus() 
{
    if ( m_pEdit && m_pEdit->GetSafeHwnd() )
        m_pEdit->SetSel( 0, 0 );
    if ( m_pList && m_pList->GetSafeHwnd() )
        m_pList->ShowWindow( SW_HIDE );
    if ( m_pButton && m_pButton->GetSafeHwnd() )
        m_pButton->BringWindowToTop();
}

void CPropertyBox::MeasureItem(LPMEASUREITEMSTRUCT 
    lpMeasureItemStruct) 
{
    // Calculate the desired height of an item based on the font
    CDC* pDC = GetDC();
    if ( pDC == NULL ) return;
    if ( m_pMeasureFont == NULL )
        m_pMeasureFont = GetParent()->GetFont();
    CFont* pOldFont = (CFont*)pDC->SelectObject( m_pMeasureFont );
    int nHeight = pDC->DrawText( "H\0", -1, CRect(), 
        DT_LEFT | DT_SINGLELINE | DT_CALCRECT );
    lpMeasureItemStruct->itemHeight = nHeight + (nHeight / 4);
    pDC->SelectObject( pOldFont );
}

void CPropertyBox::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct) 
{
    CDC* pDC = CDC::FromHandle( lpDrawItemStruct->hDC );
    if ( pDC == NULL ) return;
    
    // Default to system colors
    COLORREF crText = ::GetSysColor(COLOR_WINDOWTEXT);
    COLORREF crSelText = ::GetSysColor(COLOR_HIGHLIGHTTEXT);
    COLORREF crReadOnlyText = ::GetSysColor(COLOR_GRAYTEXT);
    COLORREF crGridLines = ::GetSysColor( COLOR_3DHILIGHT );
    COLORREF crFill = ::GetSysColor( COLOR_WINDOW );
    COLORREF crSelFill = ::GetSysColor( COLOR_HIGHLIGHT );
    COLORREF clrTextSave, clrBkSave;
    CRect rectCalc, rectStart;
    CPen penGrid( PS_SOLID, 1, crGridLines );
    CBrush brText( crFill );	
    
    // Get the property item to be drawn
    CPropertyItem* pItem = GetProperty(lpDrawItemStruct->itemID);
    rectStart = lpDrawItemStruct->rcItem;
    rectCalc = rectStart;
    rectCalc.right = rectCalc.left + m_nPropertyWidth;
    
    // Is this item selected?
    bool bSelected = ((lpDrawItemStruct->itemAction && ODA_SELECT)
        && (lpDrawItemStruct->itemState && ODS_SELECTED));
    
    // Draw the property name
    if ( bSelected )
    {
        clrTextSave = pDC->SetTextColor( crSelText );
        clrBkSave = pDC->SetBkColor( crSelFill );
        pDC->FillRect(rectCalc, &CBrush(crSelFill));
    }
    else
    {
        clrTextSave = pDC->SetTextColor( crText );
        clrBkSave = pDC->SetBkColor( crFill );
        pDC->FillRect(rectStart, &brText );
    }
    
    if ( pItem == NULL ) return;
    
    rectCalc.left += kValueMargin;
    rectCalc.right -= kValueMargin;
    
    DWORD dw = DT_LEFT | DT_SINGLELINE | DT_NOPREFIX;
    pDC->DrawText( pItem->m_strProperty, rectCalc, dw );
    m_rectValue = lpDrawItemStruct->rcItem;
    m_rectValue.left = rectCalc.right + ( 2 * kValueMargin);
    
    // Draw the property value
    if ( bSelected )
    {
        pDC->SetTextColor(crReadOnlyText);
        pDC->SetBkColor(crFill);
        m_rectValue.bottom--;
        if ( pItem->m_nFlags & PROP_READONLY )
            pDC->DrawText( pItem->m_strValue, m_rectValue, dw );
    }
    else
    {
        if ( pItem->m_nFlags & PROP_READONLY )
            clrTextSave = pDC->SetTextColor(crReadOnlyText);
        pDC->DrawText( pItem->m_strValue, m_rectValue, dw );
        if ( pItem->m_nFlags & PROP_READONLY )
            pDC->SetTextColor(clrTextSave);
    }
    
    // Draw the grid lines
    CPen* pPenTemp = (CPen*)pDC->SelectObject( &penGrid );
    pDC->MoveTo( rectStart.left - 1, rectStart.bottom - 1 );
    pDC->LineTo( rectStart.right + 1, rectStart.bottom - 1 );
    pDC->MoveTo( rectCalc.right + 1, rectCalc.bottom - 1 );
    pDC->LineTo( rectCalc.right + 1, rectCalc.top - 1 );
    pDC->SelectObject( pPenTemp );
    pDC->SetTextColor(clrTextSave);
    pDC->SetBkColor(clrBkSave);
}

void CPropertyBox::OnSize(UINT nType, int cx, int cy) 
{
    // Recalculate the column widths based on the overall size
    // and the length of the longest property name
    CListBox::OnSize(nType, cx, cy);	
    m_nPropertyWidth = cx / 2;
    
    CDC* pDC = GetDC();
    if ( pDC == NULL ) return;
    CString strName = m_strLongestPropName + "  ";
    CRect rect(0,0,0,0);
    
    CFont* pOldFont = (CFont*)pDC->SelectObject( GetFont() );
    pDC->DrawText( strName, rect, DT_CALCRECT );
    
    if ( m_nPropertyWidth > rect.right ) 
        m_nPropertyWidth = rect.right;
    pDC->SelectObject( pOldFont );
    InvalidateRect( NULL, false );
    OnSelchange();
}

BOOL CPropertyBox::OnCommand(WPARAM wParam, LPARAM lParam) 
{    
    HWND hwndCtl = (HWND) lParam;
    UINT uNotifyCode = HIWORD(wParam);
    CString strTemp;
    CPropertyItem* pItem = GetProperty( GetCurSel() );
    if ( pItem == NULL )
        return( CListBox::OnCommand( wParam, lParam ) );
    switch ( uNotifyCode )
    {
        // A child button was clicked; toggle the drop down
        case BN_CLICKED:
        {
            if ( ( hwndCtl != m_pButton->GetSafeHwnd() ) &&
                 ( hwndCtl != m_pODButton->GetSafeHwnd() ) ) 
                 break;
            ToggleList( pItem );
        }
        break;
        
        // The child list was double clicked; set the value
        case LBN_DBLCLK:
        {
            if ( hwndCtl != m_pList->GetSafeHwnd() ) break;
            int nCurSel = m_pList->GetCurSel();
            ASSERT( nCurSel != -1 );
            m_pList->GetText( nCurSel, strTemp );
            m_pList->ShowWindow( SW_HIDE );
            SetValue( pItem->m_strProperty, strTemp );
            SetFocus();
            return( 1 );
        }
        break;
        
        // The enter key was pressed; set the value
        case PN_ENTERVAL:
		{
            CString strOldValue = pItem->m_strValue;
            int nOldIndex = pItem->m_nValueIndex;
            int nCurSel = -1;
            if ( hwndCtl == m_pEdit->GetSafeHwnd() )
            {
                m_pEdit->GetWindowText( strTemp );
            }
            else
            {
                nCurSel = m_pList->GetCurSel();
                if ( nCurSel == -1 ) return( false );
                m_pList->GetText( nCurSel, strTemp );
                m_pEdit->SetWindowText( strTemp );
                m_pList->ShowWindow( SW_HIDE );
            }
            pItem->m_strValue = strTemp;
            pItem->m_nValueIndex = nCurSel;
            if ( pItem->m_nFlags & PROP_GETCHANGE )
            {
                // Notify of the property value change
                if ( SendPropChange( pItem ) )
                {
                    pItem->m_strValue = strOldValue;
                    pItem->m_nValueIndex = nOldIndex;
                    m_pEdit->SetWindowText( strOldValue );
                    return( false );
                }
            }
            SetFocus();
            return( 1 );
        }
        break;
        
        // The escape key was pressed; undo the changes
        case PN_ESCVAL:
        {
            if ( hwndCtl == m_pEdit->GetSafeHwnd() )
                m_pEdit->SetWindowText( pItem->m_strValue );
            else
            {
                m_pList->SelectString( -1, pItem->m_strValue );
                m_pList->ShowWindow( SW_HIDE );
            }
            SetFocus();
            return( 1 );
        }
        break;
        
        // Scroll up or down the list of values
        case PN_UPVAL:      case PN_DOWNVAL:
        {
            if ( hwndCtl != m_pEdit->GetSafeHwnd() ) break;
            if ( m_pList && m_pList->GetCount() > 0 )
            {
                m_pEdit->GetWindowText( strTemp );			
                int nId = m_pList->SelectString( -1, 
                    strTemp );
                if ( uNotifyCode == PN_DOWNVAL )
                    if ( ++nId >= m_pList->GetCount() - 1 )
                        nId = m_pList->GetCount() - 1;
                if ( uNotifyCode == PN_UPVAL )
                    if ( --nId < 0 ) nId = 0;
                m_pList->GetText( nId, strTemp );
                if ( GetCurSel() >= 0 )
                {
                    SetValue( pItem->m_strProperty, strTemp );
                    m_pEdit->SetWindowText( strTemp );
                    m_pEdit->SetSel( 0, -1 );
                }
            }
            return( 1 );
        }
        break;
    }
    return( CListBox::OnCommand( wParam, lParam ) );
}

BOOL CPropertyBox::PreTranslateMessage(MSG* pMsg) 
{
    switch(pMsg->message)
    {
        // Handle normal keys for control navigation
        case WM_KEYDOWN:
        {
            if ( pMsg->hwnd == m_hWnd )
                if ( ( pMsg->wParam == VK_RETURN ) || 
                     ( pMsg->wParam == VK_ESCAPE ) )
                    return( TRUE );
            if ( ( m_bTabWithinBox ) && 
                 ( pMsg->wParam == VK_TAB ) && 
                 ( GetCurSel() != -1 ) )
            {
                if ( m_pEdit && m_pEdit->GetSafeHwnd() )
                    if ( CWnd::GetFocus() == m_pEdit )
                        SetFocus();
                    else
                    {
                        m_pEdit->SetSel( 0, -1 );
                        m_pEdit->SetFocus();
                    }
                    return( TRUE );
            }
        }
        break;
        
        // Handle the ALT key for control navigation
        case WM_SYSKEYDOWN:
        {            
            if ( pMsg->wParam == VK_MENU ) 
            {
                CWnd* pFocus = GetFocus();
                if ( ( pFocus == m_pEdit ) || 
                     ( pFocus == m_pList ) )
                     return( TRUE );
                m_pEdit->SetFocus();
                return( TRUE );
            }
            else
            {
                if ( ( pMsg->wParam == VK_DOWN ) ||
                     ( pMsg->wParam == VK_UP ) )
                {                  
                    CPropertyItem* pItem = GetProperty(
                        GetCurSel() );
                    if ( pItem == NULL ) return( TRUE );
                    ToggleList( pItem );
                }
            }
            return( TRUE );
        }
        break;
    }
    return( CListBox::PreTranslateMessage(pMsg) );
}

LRESULT APIENTRY CPropertyBox::ChildSubclassProc(HWND hwnd, 
    UINT uMsg, WPARAM wParam, LPARAM lParam)
{     
    WNDPROC poldproc =  (WNDPROC)::GetWindowLong( hwnd, 
        GWL_USERDATA );
    bool bSendParentMsg = true;
    bool bHandledMsg = false;
    UINT uParentMsg = 0;
    int nId = ::GetDlgCtrlID( hwnd );
    DWORD dwEdit = ::GetWindowLong( ::GetParent( hwnd ), 
        GWL_USERDATA );
    
    switch ( uMsg )
    {
        case WM_GETDLGCODE:
            return(DLGC_WANTALLKEYS | CallWindowProc( poldproc, 
                hwnd, uMsg, wParam, lParam));
        break;        
        
        // If a child window loses focus, commit the value
        case WM_KILLFOCUS:
            uParentMsg = PN_ENTERVAL;
        break;
        
        // Make the edit window read-only if the property is
        case WM_KEYUP:  case WM_CHAR:
        case WM_CUT:    case WM_COPY:   case WM_PASTE:
        {
            if ( ( nId == PB_EDIT_ID ) && 
                 ( dwEdit == kAllowEdit ) )
            {
                bHandledMsg = true;
                bSendParentMsg = false;
            }
        }
        break;
        
        // Convert keystrokes to PN_* messages
        case WM_KEYDOWN:
        {
            bHandledMsg = true;
		    switch( wParam )
		    {
			    case VK_RETURN:
                    uParentMsg = PN_ENTERVAL;
			    break;
			    case VK_ESCAPE:
                    uParentMsg = PN_ESCVAL;
			    break;
			    case VK_DOWN:
                    if ( nId == PB_EDIT_ID )
                    {
                        uParentMsg = PN_DOWNVAL;
                        break;
                    }
                    bSendParentMsg = false;
                    bHandledMsg = false;
			    break;
			    case VK_UP:
                    if ( nId == PB_EDIT_ID )
                    {
                        uParentMsg = PN_UPVAL;
                        break;
                    }
                    bSendParentMsg = false;
                    bHandledMsg = false;
			    break;
			    default:
                    bSendParentMsg = false;
                    bHandledMsg = false;
                    if ( ( nId == PB_EDIT_ID ) && 
                         ( dwEdit == kAllowEdit) )
                        bHandledMsg = true;
                break;
            }
        }
        break;
        
        default:
            bSendParentMsg = false;
        break;
    }
    // Send the parent a PN_* message if necessary    
    if ( bSendParentMsg )
        ::SendMessage( ::GetParent( hwnd ), WM_COMMAND, 
            MAKEWPARAM(nId, uParentMsg), (LPARAM)hwnd);
    
    // If we handle it, return 0;  otherwise use the old WNDPROC
    if ( bHandledMsg ) return( 0 );
    return CallWindowProc(poldproc, hwnd, uMsg, wParam, lParam); 
}

void CPropertyBox::HideChildren( void )
{    
    m_pEdit->ShowWindow( SW_HIDE );
    m_pButton->ShowWindow( SW_HIDE );
    m_pODButton->ShowWindow( SW_HIDE );
    if ( m_pList ) m_pList->ShowWindow( SW_HIDE );
    m_pList = NULL;
}

bool CPropertyBox::SendNotification( UINT nCode,
    CPropertyItem*& pItem, const RECT& rect )
{
    CWnd* pParent = GetParent();
    if ( pParent == NULL ) return( false );
    
    // Build the notification structure and send the message
    LPNMPBDISPINFO lpnmpb = new NMPBDISPINFO;
    memset( lpnmpb, 0, sizeof( NMPBDISPINFO ) );
    lpnmpb->hwndFrom = m_hWnd;
    lpnmpb->idFrom = GetDlgCtrlID();
    lpnmpb->code = nCode;
    lpnmpb->pItem = pItem;
    memcpy( &(lpnmpb->rectOwnerDrop), &rect, sizeof(RECT) );
    
    bool bRet= ( pParent->SendMessage( WM_NOTIFY, 
        (WPARAM)(GetDlgCtrlID()),
        (LPARAM)lpnmpb ) == 0 ? false : true );
    delete lpnmpb;
    return( bRet );
}

void CPropertyBox::ToggleList( CPropertyItem* pItem)
{
    CRect rect;
    GetItemRect( GetCurSel(), rect );
    int nItemHeight = rect.Height() * 4;
    rect.top = rect.bottom;
    rect.bottom = rect.top + nItemHeight;
    rect.left += m_nPropertyWidth;
    
    if ( pItem->m_nFlags & PROP_DROPLIST )
    {
        // If the list is visible, hide it and commit the value
        if ( m_pList && m_pList->IsWindowVisible() )
        {
            int nCurSel = m_pList->GetCurSel();
            if ( nCurSel >= 0 )
            {
                m_pList->GetText(nCurSel, pItem->m_strValue);
                m_pEdit->SetWindowText(pItem->m_strValue);
            }
            m_pList->ShowWindow( SW_HIDE );
            if ( m_pEdit->GetSafeHwnd() && 
                m_pEdit->IsWindowVisible() ) 
                    m_pEdit->SetFocus();
        }
        else
        {
            // Display the list selecting the current value
            // if possible
            m_pEdit->SetSel( 0, -1 );
            SetFocus();
            nItemHeight = m_pList->GetItemHeight( 0 )
                * ( m_pList->GetCount() + 1 );
            rect.bottom = rect.top + nItemHeight;
            m_pList->MoveWindow( rect );
            m_pList->SetFont( GetFont() );
            m_pList->ShowWindow( SW_SHOW );
            if ( m_pList->SelectString( -1, 
                pItem->m_strValue ) == -1 )
                    m_pList->SetCurSel( -1 );
            m_pList->SetFocus();
        }
    }
    else
        // The application must handle the drop down
        if ( pItem->m_nFlags & PROP_OWNERDROP )
            SendOwnerDrop( pItem, rect );
}

UINT CPropertyBox::OnGetDlgCode() 
{
    // The return value indicates whether this window wants to 
    // process TAB keystrokes based on the m_bTabWithinBox flag
    UINT nCode = CListBox::OnGetDlgCode();
    if ( m_bTabWithinBox ) nCode |= DLGC_WANTTAB;
    return( nCode );
}

int CPropertyBox::AddProperty( CPropertyItem* pItem )
{
    if ( pItem == NULL ) return( -1 );
    CPropertyItem* pLookup = GetProperty( pItem->m_strProperty );
    
    // Each property name can only appear once
    if ( pLookup ) return( LB_ERR );
    if ( pItem->m_nFlags & PROP_DROPLIST )
    {
        if ( ( pItem->m_nFlags & PROP_OWNERDROP ) ||
            ( pItem->m_nFlags & PROP_READONLY ) ) return( -1 );
        
        // If the property has a value list box, subclass it
        if ( pItem->m_pLB )
        {
            WNDPROC poldproc = (WNDPROC)::GetWindowLong( 
                pItem->m_pLB->GetSafeHwnd(), GWL_WNDPROC );
            if ( poldproc != (WNDPROC) ChildSubclassProc )
            {
                SetWindowLong( pItem->m_pLB->GetSafeHwnd(), 
                    GWL_USERDATA, (LONG) poldproc );
                SetWindowLong( pItem->m_pLB->GetSafeHwnd(),
                    GWL_WNDPROC, (LONG) ChildSubclassProc );
            }
        }
    }
    
    // Check for invalid flag combinations
    if ( pItem->m_nFlags & PROP_OWNERDROP )
        if ( ( pItem->m_nFlags & PROP_DROPLIST ) ||
        ( pItem->m_nFlags & PROP_READONLY ) ) return( -1 );
    if ( pItem->m_nFlags & PROP_READONLY )
        if ( ( pItem->m_nFlags & PROP_DROPLIST ) ||
            ( pItem->m_nFlags & PROP_OWNERDROP ) ) return( -1 );
    
    m_mapPropToData.SetAt( pItem->m_strProperty, pItem );
    if ( m_strLongestPropName.GetLength() < 
        (int)strlen( pItem->m_strProperty ) )
    {
        m_strLongestPropName = pItem->m_strProperty;
        CRect rect;
        GetClientRect( rect );
        SendMessage( WM_SIZE, (WPARAM)SIZE_RESTORED, 
            MAKELPARAM(rect.Width(), rect.Height()) );
    }
    int nRet = AddString( pItem->m_strProperty );
    OnSelchange();
    
    return( nRet );
}

int CPropertyBox::DeleteProperty( LPCTSTR lpszProperty )
{
    int nId = FindString( -1, lpszProperty );
    if ( nId != -1 )
    {
        HideChildren();
        nId = DeleteString( nId );
        CPropertyItem* pItem = GetProperty( lpszProperty );
        if ( pItem ) m_mapPropToData.RemoveKey( lpszProperty );
        delete pItem;
    }
    InvalidateRect( NULL );
    SetCurSel( -1 );
    return( nId );
}

bool CPropertyBox::SetValue(LPCTSTR lpszProperty, 
    LPCTSTR lpszValue)
{
    CPropertyItem* pItem = GetProperty( lpszProperty );
    if ( pItem )
    {				
        if ( pItem == NULL ) return( false );
        CString strOldValue = pItem->m_strValue;
        pItem->m_strValue = lpszValue;
        if ( pItem->m_nFlags & PROP_GETCHANGE )
        {
            // Notify of the property value change
            if ( SendPropChange( pItem ) )
            {
                pItem->m_strValue = strOldValue;
                return( false );
            }
        }
        int nId = FindString( -1, lpszProperty );
        if ( nId == GetCurSel() )
            m_pEdit->SetWindowText( lpszValue );

        CRect rectItem;
        GetItemRect( nId, rectItem );
        InvalidateRect( rectItem );
        
        return( true );
    }
    return( false );
}

bool CPropertyBox::GetValue( LPCTSTR lpszProperty, 
    CString& strValue)
{
    CPropertyItem* pItem = GetProperty( lpszProperty );
    if ( pItem == NULL ) return( false );
    strValue = pItem->m_strValue;
    return( true );
}

void CPropertyBox::SelectProperty( const int& nIndex )
{
    if ( ( nIndex < -1 ) || ( nIndex >= GetCount() ) ) 
        return;
    SetCurSel( nIndex );
    InvalidateRect( NULL );
    OnSelchange();
}

CPropertyItem* CPropertyBox::GetProperty( LPCTSTR lpszProperty )
{
    CPropertyItem* pItem = NULL;
    m_mapPropToData.Lookup( lpszProperty, (void*&)pItem );
    return( pItem );
}

CPropertyItem* CPropertyBox::GetProperty( const int& nIndex )
{
    if ( ( nIndex < 0 ) || ( nIndex >= GetCount() ) ) 
        return( false );
    CString strProp;
    GetText( nIndex, strProp );
    CPropertyItem* pItem = GetProperty( strProp );
    return( pItem );
}

void CPropertyBox::RemoveAll()
{
    POSITION pos = m_mapPropToData.GetStartPosition();
    while ( pos != NULL )
    {
        CPropertyItem* pItem = NULL;
        CString strProp;
        m_mapPropToData.GetNextAssoc(pos, strProp, (void*&)pItem);
        delete pItem;
    }
    m_mapPropToData.RemoveAll();
    if ( GetSafeHwnd() ) ResetContent();
}
